/**
 * Module dependencies
 */

var http = require('http')
  , fs = require('fs')
  , path = require('path')
  , join = path.join
  , configure = require('./configure')
  , _ = require('./_')

  , _index = 0
  , _middleware
  , _out
  , _req
  , _res;

function next(err) {
  var layer = _middleware[_index]
    , arity;

  _index += 1;

  // completed middleware stack
  if (!layer) {
    if (_out) _out(err);
    if (err) {
      _res.end();
    } else if (_res.headersSent) {
      _res.end();
    } else {
      _res.statusCode = 404;
      _res.setHeader('Content-Type', 'text/html');
      _res.end('404 not found');
    }
    return;
  }

  arity = layer.handle.length;

  if (err && arity === 4) {
    layer.handle(err, _req, _res, next);
  } else if (err && arity < 4) {
    next(err);
  } else if (!err && arity < 4) {
    layer.handle(_req, _res, next);
  } else {
    next();
  }
}

var application = function application(conf) {

  this._root = process.env.PWD;

  this.config = configure.user(conf).config();

  this.controllers = (function () {
    var controllers = fs.readdirSync(join(this._root, this.config.paths.controllers))
      , obj = {};

    _.each(controllers, function (controller) {
      var name = path.basename(controller, '.js');
      try {
        obj[name] = require(join(this._root, this.config.paths.controllers, controller));
      } catch (err) {
        throw err;
      }
    }.bind(this));
    return obj;

  }.bind(this)());

  this.middleware = _middleware = (function () {
    var stack = [];
    _.each(this.config.middleware, function (layer) {
      var name = Object.keys(layer)[0]
        , mw;
      try {
        mw = require('./middleware/' + name);
        stack.push({
          handle: mw(layer[name].options)
        });
      } catch (err) {
        throw err;
      }
    });
    return stack;
  }.bind(this)());

  this.routes = (function () {
    var routes;
    try {
      var routesPath = join(this._root, 'config', 'routes');
      if (this.config.paths.routes) {
        routesPath = join(this._root, this.config.paths.routes);
      }
      routes = require(routesPath);
    } catch (err) {
      throw err;
    }
    return routes;
  }.bind(this)());

  if (application.caller !== application.getInstance) {
    throw new Error('This object cannot be instantiated.');
  }
};

application.instance = null;

application.getInstance = function (conf) {
  if (this.instance === null || this.instance === undefined) {
    this.instance = new application(conf);
  }
  return this.instance;
};

application.prototype.handle = function (req, res, out) {
  _index = 0;
  _req = req;
  _res = res;
  _out = out;
  next();
};

application.prototype.listen = function () {
  var server = http.createServer(this.handle);
  return server.listen.apply(server, arguments);
};

exports = module.exports = application.getInstance;
